@@ -167,6 +167,10 @@
  *      text = "item1"
  *      ts = 1440801895 # timestamp (optional)
  *  } ]
+ *  menu-cmd-hist = [ {
+ *      text = "item1"
+ *      ts = 1440801895 # timestamp (optional)
+ *  } ]
  *  active-gtab = 0
  *  use-term-multiplexer = true
  *  color-scheme = "almost-default"
@@ -681,6 +685,7 @@ load_state(JSON_Object *root, int reread)
 	load_history(root, "search-hist", &curr_stats.search_hist);
 	load_history(root, "prompt-hist", &curr_stats.prompt_hist);
 	load_history(root, "lfilt-hist", &curr_stats.filter_hist);
+	load_history(root, "menu-cmd-hist", &curr_stats.menucmd_hist);
 }
 
 /* Loads global tabs from JSON. */
@@ -709,9 +714,17 @@ load_gtabs(JSON_Object *root, int reread)
 		}
 	}
 
-	int i, n;
+	int gtab_count = json_array_get_count(gtabs);
+	if(gtab_count > 1)
+	{
+		/* More than one global tab might require updating tab scope option. */
+		cfg.pane_tabs = 0;
+		load_tabscope_option();
+	}
+
+	int i;
 	view_t *left = &lwin, *right = &rwin;
-	for(i = 0, n = json_array_get_count(gtabs); i < n; ++i)
+	for(i = 0; i < gtab_count; ++i)
 	{
 		JSON_Object *gtab = json_array_get_object(gtabs, i);
 
@@ -1033,7 +1046,7 @@ load_cmds(JSON_Object *root)
 			char *cmdadd_cmd = format_str("command %s %s", name, cmd);
 			if(cmdadd_cmd != NULL)
 			{
-				exec_commands(cmdadd_cmd, curr_view, CIT_COMMAND);
+				cmds_dispatch(cmdadd_cmd, curr_view, CIT_COMMAND);
 				free(cmdadd_cmd);
 			}
 		}
@@ -1429,6 +1442,11 @@ serialize_state(int vinfo)
 		store_history(root, "lfilt-hist", &curr_stats.filter_hist);
 	}
 
+	if(vinfo & VINFO_MCHISTORY)
+	{
+		store_history(root, "menu-cmd-hist", &curr_stats.menucmd_hist);
+	}
+
 	if(vinfo & VINFO_REGISTERS)
 	{
 		store_regs(root);
@@ -1507,6 +1525,11 @@ merge_states(int vinfo, int session_load, JSON_Object *current,
 		merge_history(session_load, current, admixture, "lfilt-hist");
 	}
 
+	if(vinfo & VINFO_MCHISTORY)
+	{
+		merge_history(session_load, current, admixture, "menu-cmd-hist");
+	}
+
 	if(vinfo & VINFO_REGISTERS)
 	{
 		merge_regs(current, admixture);
@@ -1761,7 +1784,8 @@ timestamp_cmp(const void *a, const void *b)
 	double lhs_ts = -1, rhs_ts = -1;
 	get_double(*(const JSON_Object **)a, "ts", &lhs_ts);
 	get_double(*(const JSON_Object **)b, "ts", &rhs_ts);
-	return lhs_ts - rhs_ts;
+	/* There is no fractional part, so can compare similar to integers. */
+	return SORT_CMP(lhs_ts, rhs_ts);
 }
 
 /* Merges two lists of associations. */
@@ -2238,6 +2262,8 @@ store_global_options(JSON_Object *root)
 	append_dstr(options, format_str("histcursor=%s",
 			escape_spaces(vle_opts_get("histcursor", OPT_GLOBAL))));
 	append_dstr(options, format_str("history=%d", cfg.history_len));
+	append_dstr(options, format_str("hloptions=%s",
+			escape_spaces(vle_opts_get("hloptions", OPT_GLOBAL))));
 	append_dstr(options, format_str("%shlsearch", cfg.hl_search ? "" : "no"));
 	append_dstr(options, format_str("%siec",
 				cfg.sizefmt.ieci_prefixes ? "" : "no"));
@@ -2257,6 +2283,8 @@ store_global_options(JSON_Object *root)
 	append_dstr(options, format_str("rulerformat=%s",
 				escape_spaces(cfg.ruler_format)));
 	append_dstr(options, format_str("%srunexec", cfg.auto_execute ? "" : "no"));
+	append_dstr(options, format_str("navoptions=%s",
+				escape_spaces(vle_opts_get("navoptions", OPT_GLOBAL))));
 	append_dstr(options, format_str("previewoptions=%s",
 				escape_spaces(vle_opts_get("previewoptions", OPT_GLOBAL))));
 	append_dstr(options, format_str("%sscrollbind", cfg.scroll_bind ? "" : "no"));
@@ -2279,6 +2307,8 @@ store_global_options(JSON_Object *root)
 				cfg.sort_numbers ? "" : "no"));
 	append_dstr(options, format_str("statusline=%s",
 				escape_spaces(cfg.status_line)));
+	append_dstr(options, format_str("tabline=%s",
+				escape_spaces(vle_opts_get("tabline", OPT_GLOBAL))));
 	append_dstr(options, format_str("syncregs=%s",
 			escape_spaces(vle_opts_get("syncregs", OPT_GLOBAL))));
 	append_dstr(options, format_str("tablabel=%s",
@@ -2296,6 +2326,8 @@ store_global_options(JSON_Object *root)
 	append_dstr(options, format_str("%strash", cfg.use_trash ? "" : "no"));
 	append_dstr(options, format_str("tuioptions=%s",
 			escape_spaces(vle_opts_get("tuioptions", OPT_GLOBAL))));
+	append_dstr(options, format_str("uioptions=%s",
+			escape_spaces(vle_opts_get("uioptions", OPT_GLOBAL))));
 	append_dstr(options, format_str("undolevels=%d", cfg.undo_levels));
 	append_dstr(options, format_str("vicmd=%s%s", escape_spaces(cfg.vi_command),
 			cfg.vi_cmd_bg ? " &" : ""));
@@ -2307,6 +2339,8 @@ store_global_options(JSON_Object *root)
 			escape_spaces(vle_opts_get("confirm", OPT_GLOBAL))));
 	append_dstr(options, format_str("dotdirs=%s",
 			escape_spaces(vle_opts_get("dotdirs", OPT_GLOBAL))));
+	append_dstr(options, format_str("mouse=%s",
+			escape_spaces(vle_opts_get("mouse", OPT_GLOBAL))));
 	append_dstr(options, format_str("caseoptions=%s",
 			escape_spaces(vle_opts_get("caseoptions", OPT_GLOBAL))));
 	append_dstr(options, format_str("suggestoptions=%s",
