@@ -21,8 +21,7 @@
 
 #include <regex.h>
 
-#include <sys/stat.h> /* stat umask() */
-#include <sys/types.h> /* mode_t */
+#include <sys/stat.h> /* stat */
 #include <fcntl.h>
 #include <unistd.h> /* unlink() */
 
@@ -33,9 +32,10 @@
 
 #include <assert.h> /* assert() */
 #include <errno.h> /* errno */
+#include <math.h> /* fabsf() sqrtf() */
 #include <stddef.h> /* NULL size_t */
 #include <stdint.h> /* uint64_t */
-#include <stdio.h> /* snprintf() */
+#include <stdio.h> /* FILE snprintf() */
 #include <stdlib.h> /* free() malloc() */
 #include <string.h> /* memset() strcat() strcmp() strdup() strlen() */
 #include <time.h> /* clock_gettime() */
@@ -87,6 +87,21 @@
 /* Key used to switch to progress dialog. */
 #define IO_DETAILS_KEY 'i'
 
+/* Number of entries in history windows used to compute ETA. */
+#define ETA_HISTORY_SIZE 10
+
+/* A circular buffer for computing weighted average. */
+typedef struct
+{
+	/* We're only adding here, so no need to keep head index, because it's zero
+	 * until wrapping and `pos` after wrapping. */
+	int wrapped; /* Whether the buffer has been wrapped. */
+	int pos;     /* Next position for writing (tail). */
+
+	uint64_t entries[ETA_HISTORY_SIZE]; /* Buffer's data. */
+}
+history_window_t;
+
 /* Object for auxiliary information related to progress of operations in
  * io_progress_changed() handler. */
 typedef struct
@@ -107,10 +122,15 @@ typedef struct
 	int progress_bar_max;   /* Width of progress bar during previous call. */
 
 	/* State of rate calculation. */
+	long long start_time;     /* Time of starting the operation. */
 	long long last_calc_time; /* Time of last rate calculation. */
 	uint64_t last_seen_byte;  /* Position at the time of last call. */
-	uint64_t rate;            /* Rate in bytes per millisecond. */
+	float last_rate;          /* Rate in bytes per millisecond. */
+	int last_eta;             /* Last reported ETA. */
 	char *rate_str;           /* Rate formatted as a string. */
+	char *eta_str;            /* ETA formatted as a string. */
+	history_window_t speed_window; /* Smoothing of speed. */
+	history_window_t eta_window;   /* Smoothing of ETA. */
 
 	/* Whether progress is displayed in a dialog, rather than on status bar. */
 	int dialog;
@@ -119,9 +139,18 @@ typedef struct
 }
 progress_data_t;
 
+/* Type of function that returns next weight for weighted average. */
+typedef float (*weight_f)(int idx, float last_weight);
+
 static void io_progress_changed(const io_progress_t *state);
 static int calc_io_progress(const io_progress_t *state, int *skip);
-static void update_io_rate(progress_data_t *pdata, const ioeta_estim_t *estim);
+static void update_io_stats(progress_data_t *pdata, const ioeta_estim_t *estim);
+static void window_add(history_window_t *win, uint64_t value);
+static float window_average(const history_window_t *win, weight_f weight);
+static float linear_weight(int idx, float last_weight);
+static float doubling_weight(int idx, float last_weight);
+static void format_io_stats(progress_data_t *pdata, long long now_ms,
+		uint64_t imm_rate);
 static void update_progress_bar(progress_data_t *pdata,
 		const ioeta_estim_t *estim);
 static void io_progress_fg(const io_progress_t *state, int progress);
@@ -251,28 +280,165 @@ calc_io_progress(const io_progress_t *state, int *skip)
 
 /* Updates rate of operation. */
 static void
-update_io_rate(progress_data_t *pdata, const ioeta_estim_t *estim)
+update_io_stats(progress_data_t *pdata, const ioeta_estim_t *estim)
 {
 	long long current_time_ms = time_in_ms();
 	long long elapsed_time_ms = current_time_ms - pdata->last_calc_time;
 
 	if(elapsed_time_ms == 0 ||
-			(pdata->last_seen_byte != 0 && elapsed_time_ms < 3000))
+			(pdata->last_seen_byte != 0 && elapsed_time_ms < 1000))
 	{
-		/* Rate is updated initially and then once in 3000 milliseconds. */
+		/* Rate is updated initially and then once in 1000 milliseconds. */
 		return;
 	}
 
 	uint64_t bytes_difference = estim->current_byte - pdata->last_seen_byte;
-	pdata->rate = bytes_difference/elapsed_time_ms;
+	uint64_t imm_rate = bytes_difference/elapsed_time_ms;
+
+	/* Smooth immediate rate. */
+	window_add(&pdata->speed_window, imm_rate);
+	float rate = window_average(&pdata->speed_window, &doubling_weight);
+
+	/* Second round of adjusting speed to not deviate from previous value too
+	 * much. */
+	if(rate >= 1 && pdata->last_rate >= 1)
+	{
+		float minV = MIN(rate, pdata->last_rate);
+		float maxV = MAX(rate, pdata->last_rate);
+		rate = pdata->last_rate + (rate - pdata->last_rate)*minV/maxV;
+	}
+
+	/* Acceleration. */
+	float a = (rate - pdata->last_rate)/elapsed_time_ms;
+	/* Remains this much. */
+	uint64_t r = estim->total_bytes - estim->current_byte;
+
+	int eta;
+	if(fabsf(a) >= 1 && rate*rate + 2*a*r >= 0)
+	{
+		eta = (sqrtf(rate*rate + 2*a*r) - rate)/a;
+	}
+	else if(fabsf(rate) >= 1)
+	{
+		eta = r/rate;
+	}
+	else
+	{
+		eta = pdata->last_eta;
+	}
+
+	/* Smooth ETA rate. */
+	window_add(&pdata->eta_window, eta);
+	float smooth_eta = window_average(&pdata->eta_window, &linear_weight);
+
 	pdata->last_calc_time = current_time_ms;
 	pdata->last_seen_byte = estim->current_byte;
+	pdata->last_rate = rate;
+	pdata->last_eta = smooth_eta;
 
+	format_io_stats(pdata, current_time_ms, imm_rate);
+}
+
+/* Adds an entry to window. */
+static void
+window_add(history_window_t *win, uint64_t value)
+{
+	win->entries[win->pos++] = value;
+	if(win->pos == ETA_HISTORY_SIZE)
+	{
+		win->pos = 0;
+		win->wrapped = 1;
+	}
+}
+
+/* Computes weighted average on entries of the passed in window.  Returns the
+ * average. */
+static float
+window_average(const history_window_t *win, weight_f weight)
+{
+	assert((win->pos != 0 || win->wrapped) && "Window must not be empty.");
+
+	float average = 0.0f;
+
+	/* Current weight. */
+	float w = 0.0f;
+	/* Total weight. */
+	float ws = 0.0f;
+
+	/* Start positions within circular buffer. */
+	int pos = (win->wrapped ? win->pos : 0);
+	/* Number of used positions. */
+	int count = (win->wrapped ? ETA_HISTORY_SIZE : win->pos);
+
+	int i;
+	for(i = 0; i < count; ++i)
+	{
+		w = weight(i, w);
+		ws += w;
+
+		average += w*win->entries[pos];
+		pos = (pos + 1)%ETA_HISTORY_SIZE;
+	}
+
+	average /= ws;
+	return average;
+}
+
+/* Linear weight function (1, 2, 3, ...).  Returns the next weight. */
+static float
+linear_weight(int idx, float last_weight)
+{
+	return 1 + idx;
+}
+
+/* Weight function with the factor of 2 (1, 2, 4, ...).  Returns the next
+ * weight. */
+static float
+doubling_weight(int idx, float last_weight)
+{
+	return (idx == 0 ? 1.0f : last_weight*2.0f);
+}
+
+/* Updates pdata->*_str fields. */
+static void
+format_io_stats(progress_data_t *pdata, long long now_ms, uint64_t imm_rate)
+{
 	char rate_str[64];
-	(void)friendly_size_notation(pdata->rate*1000, sizeof(rate_str) - 8,
-			rate_str);
+	(void)friendly_size_notation(imm_rate*1000, sizeof(rate_str) - 8, rate_str);
 	strcat(rate_str, "/s");
 	replace_string(&pdata->rate_str, rate_str);
+
+	/* Do not show ETA for the first 5 seconds.  Really short operations need no
+	 * ETA and long ones might have incorrect one at first. */
+	if(now_ms - pdata->start_time < 5000)
+	{
+		return;
+	}
+
+	enum
+	{
+		Minute = 60,
+		Hour = 60*Minute,
+		Day = 24*Hour,
+	};
+
+	int eta_s = ceil(pdata->last_eta/1000.0);
+	int d = eta_s/Day;
+	int h = eta_s%Day/Hour;
+	int m = eta_s%Hour/Minute;
+	int s = eta_s%Minute;
+
+	char *time_str;
+	if(d > 0)
+	{
+		time_str = format_str("%dd %02d:%02d:%02d", d, h, m, s);
+	}
+	else
+	{
+		time_str = format_str("%02d:%02d:%02d", h, m, s);
+	}
+	put_string(&pdata->eta_str, format_str("~%s left", time_str));
+	free(time_str);
 }
 
 /* Updates progress bar of operation. */
@@ -366,7 +532,7 @@ io_progress_fg(const io_progress_t *state, int progress)
 
 	item_num = MIN(estim->current_item + 1, estim->total_items);
 
-	update_io_rate(pdata, estim);
+	update_io_stats(pdata, estim);
 
 	if(progress < 0)
 	{
@@ -387,13 +553,14 @@ io_progress_fg(const io_progress_t *state, int progress)
 
 		draw_msgf(title, ctrl_msg, pdata->width,
 				"Location: %s\nItem:     %d of %" PRINTF_ULL "\n"
-				"Overall:  %s/%s (%2d%%) %s\n"
+				"Overall:  %5s/%-5s (%d%%)  |  %s  %s  %s\n"
 				"%s\n"
 				" \n" /* Space is on purpose to preserve empty line. */
 				"file %s\nfrom %s%s%s",
 				replace_home_part(ops->target_dir), item_num,
 				(unsigned long long)estim->total_items, current_size_str,
 				total_size_str, progress/IO_PRECISION, pdata->rate_str,
+				pdata->eta_str[0] == '\0' ? "" : "|", pdata->eta_str,
 				pdata->progress_bar, item_name, src_path, as_part, file_progress);
 
 		free(file_progress);
@@ -522,8 +689,7 @@ format_pretty_path(const char base_dir[], const char path[], char pretty[],
 }
 
 int
-fops_is_name_list_ok(int count, int nlines, char *list[], char *files[],
-		char **error)
+fops_is_name_list_ok(int count, int nlines, char *list[], char **error)
 {
 	int i;
 
@@ -544,30 +710,6 @@ fops_is_name_list_ok(int count, int nlines, char *list[], char *files[],
 	{
 		chomp(list[i]);
 
-		if(files != NULL)
-		{
-			char *file_s = find_slashr(files[i]);
-			char *list_s = find_slashr(list[i]);
-			if(list_s != NULL || file_s != NULL)
-			{
-				if(list_s == NULL || file_s == NULL ||
-						list_s - list[i] != file_s - files[i] ||
-						strnoscmp(files[i], list[i], list_s - list[i]) != 0)
-				{
-					if(file_s == NULL)
-					{
-						put_string(error, format_str("Name \"%s\" contains slash",
-									list[i]));
-					}
-					else
-					{
-						put_string(error, format_str("Won't move \"%s\" file", files[i]));
-					}
-					return 0;
-				}
-			}
-		}
-
 		if(list[i][0] != '\0' && is_in_string_array(list, i, list[i]))
 		{
 			put_string(error, format_str("Name \"%s\" duplicates", list[i]));
@@ -601,6 +743,81 @@ fops_is_copy_list_ok(const char dst[], int count, char *list[], int force,
 	return 1;
 }
 
+int
+fops_check_moves_on_rename(int count, char *list[], char *files[], char **error)
+{
+	int moves = 0;
+
+	int i;
+	for(i = 0; i < count; ++i)
+	{
+		/* Assuming that fops_is_name_list_ok() did chomp(), so not repeating it. */
+
+		if(list[i][0] == '\0')
+		{
+			/* This file won't be renamed. */
+			continue;
+		}
+
+		char *file_s = find_slashr(files[i]);
+		char *list_s = find_slashr(list[i]);
+		if(list_s == NULL && file_s == NULL)
+		{
+			/* No prefixes, not a move. */
+			continue;
+		}
+
+		if(list_s != NULL && file_s != NULL &&
+				list_s - list[i] == file_s - files[i] &&
+				strnoscmp(files[i], list[i], list_s - list[i]) == 0)
+		{
+			/* Prefixes are identical. */
+			continue;
+		}
+
+		char *list_tail = (list_s == NULL ? list[i] : list_s);
+		char *file_tail = (file_s == NULL ? files[i] : file_s);
+
+		char list_tail_c = *list_tail;
+		char file_tail_c = *file_tail;
+		*list_tail = '\0';
+		*file_tail = '\0';
+
+		int same_prefix = paths_are_same(list[i][0] == '\0' ? "." : list[i],
+				files[i][0] == '\0' ? "." : files[i]);
+		*list_tail = list_tail_c;
+		*file_tail = file_tail_c;
+
+		if(same_prefix)
+		{
+			/* Even though prefixes look different, they point at the same path. */
+			continue;
+		}
+
+		++moves;
+
+		if(list_s == NULL)
+		{
+			/* Empty prefix is some parent directory which must already exist. */
+			continue;
+		}
+
+		char list_slash = *list_s;
+		*list_s = '\0';
+		int dir_exists = is_dir(list[i]);
+		*list_s = list_slash;
+
+		if(!dir_exists)
+		{
+			char *msg = format_str("Won't create directory for \"%s\"", list[i]);
+			put_string(error, msg);
+			return -1;
+		}
+	}
+
+	return moves;
+}
+
 int
 fops_is_rename_list_ok(char *files[], char is_dup[], int len, char *list[],
 		char **error)
@@ -619,7 +836,7 @@ fops_is_rename_list_ok(char *files[], char is_dup[], int len, char *list[],
 		int j;
 		for(j = 0; j < len; ++j)
 		{
-			if(strcmp(list[i], files[j]) == 0 && !is_dup[j])
+			if(paths_are_equal(list[i], files[j]) && !is_dup[j])
 			{
 				is_dup[j] = 1;
 				break;
@@ -744,6 +961,8 @@ fops_get_ops(OPS main_op, const char descr[], const char base_dir[],
 	if(ops->use_system_calls)
 	{
 		progress_data_t *const pdata = alloc_progress_data(ops->bg, ops);
+		pdata->dialog = cfg.always_show_io_details;
+
 		const io_cancellation_t cancellation = { .hook = &ui_cancellation_hook };
 		ops->estim = ioeta_alloc(pdata, cancellation);
 	}
@@ -889,25 +1108,22 @@ edit_list(ext_edit_t *ext_edit, size_t orig_len, char *orig[], int *edited_len,
 	*edited_len = 0;
 
 	char rename_file[PATH_MAX + 1];
-	generate_tmp_file_name("vifm.rename", rename_file, sizeof(rename_file));
-
-	strlist_t prepared = ext_edit_prepare(ext_edit, orig, orig_len);
-
-	/* Allow temporary file to be only readable and writable by current user. */
-	mode_t saved_umask = umask(~0600);
-	const int write_error = (write_file_of_lines(rename_file, prepared.items,
-				prepared.nitems) != 0);
-	(void)umask(saved_umask);
-
-	free_string_array(prepared.items, prepared.nitems);
-
-	if(write_error)
+	FILE *file = make_file_in_tmp("vifm.rename", 0600, /*auto_delete=*/0,
+			rename_file, sizeof(rename_file));
+	if(file == NULL)
 	{
 		show_error_msgf("Error Getting List Of Renames",
 				"Can't create temporary file \"%s\": %s", rename_file, strerror(errno));
 		return NULL;
 	}
 
+	strlist_t prepared = ext_edit_prepare(ext_edit, orig, orig_len);
+
+	write_lines_to_file(file, prepared.items, prepared.nitems);
+	fclose(file);
+
+	free_string_array(prepared.items, prepared.nitems);
+
 	if(vim_view_file(rename_file, -1, -1, 0) != 0)
 	{
 		unlink(rename_file);
@@ -918,11 +1134,12 @@ edit_list(ext_edit_t *ext_edit, size_t orig_len, char *orig[], int *edited_len,
 
 	int result_len;
 	char **result = read_file_of_lines(rename_file, &result_len);
+	int error = errno;
 	unlink(rename_file);
 	if(result == NULL)
 	{
 		show_error_msgf("Error Getting List Of Renames",
-				"Can't open temporary file \"%s\": %s", rename_file, strerror(errno));
+				"Can't open temporary file \"%s\": %s", rename_file, strerror(error));
 		return NULL;
 	}
 
@@ -982,10 +1199,17 @@ alloc_progress_data(int bg, void *info)
 	pdata->progress_bar_max = 0;
 
 	/* Time of starting the operation to have meaningful first rate. */
-	pdata->last_calc_time = time_in_ms();
+	pdata->start_time = time_in_ms();
+	pdata->last_calc_time = pdata->start_time;
 	pdata->last_seen_byte = 0;
-	pdata->rate = 0;
+	pdata->last_rate = 0.0f;
+	pdata->last_eta = 0;
 	pdata->rate_str = strdup("? B/s");
+	pdata->eta_str = strdup("");
+	pdata->eta_window.wrapped = 0;
+	pdata->eta_window.pos = 0;
+	pdata->speed_window.wrapped = 0;
+	pdata->speed_window.pos = 0;
 
 	pdata->dialog = 0;
 	pdata->width = 0;
@@ -1006,6 +1230,12 @@ time_in_ms(void)
 	return current_time.tv_sec*1000 + current_time.tv_nsec/1000000;
 }
 
+int
+fops_active(const ops_t *ops)
+{
+	return !ops->aborted && !ui_cancellation_requested();
+}
+
 void
 fops_free_ops(ops_t *ops)
 {
@@ -1033,6 +1263,7 @@ fops_free_ops(ops_t *ops)
 		progress_data_t *pdata = ops->estim->param;
 		free(pdata->progress_bar);
 		free(pdata->rate_str);
+		free(pdata->eta_str);
 		free(pdata);
 	}
 	ops_free(ops);
