@@ -19,8 +19,6 @@
 
 #include "running.h"
 
-#include <curses.h> /* FALSE curs_set() */
-
 #include <sys/stat.h> /* stat */
 #ifdef _WIN32
 #define WIN32_LEAN_AND_MEAN
@@ -86,10 +84,10 @@ FileHandleLink;
 
 static void handle_file(view_t *view, FileHandleExec exec,
 		FileHandleLink follow);
-static int is_runnable(view_t *view, const char full_path[], int type,
-		int force_follow);
+static int is_multiselect(view_t *view);
+static int is_runnable(const char full_path[], int type, int force_follow);
 static int is_executable(const char full_path[], const dir_entry_t *curr,
-		int dont_execute, int runnable);
+		int runnable);
 #ifdef _WIN32
 static void run_win_executable(char full_path[], int elevate);
 static int run_win_executable_as_evaluated(const char full_path[]);
@@ -99,14 +97,13 @@ static void execute_file(const char full_path[], int elevate);
 static void run_selection(view_t *view, int dont_execute);
 static void run_with_defaults(view_t *view);
 static void run_selection_separately(view_t *view, int dont_execute);
-static int is_multi_run_compat(view_t *view, const char prog_cmd[]);
+static int is_multi_run_compat(const char prog_cmd[]);
 static void run_explicit_prog(view_t *view, const char prog_spec[], int pause,
 		int force_bg);
 static void run_implicit_prog(view_t *view, const char prog_spec[], int pause,
 		int force_bg);
 static void view_current_file(const view_t *view);
 static void follow_link(view_t *view, int follow_dirs, int ultimate);
-static void enter_dir(struct view_t *view);
 static int cd_to_parent_dir(view_t *view);
 static void extract_last_path_component(const char path[], char buf[]);
 static char * run_shell_prepare(const char command[], ShellPause pause,
@@ -127,7 +124,8 @@ static int try_run_with_filetype(view_t *view, const assoc_records_t assocs,
 static void output_to_statusbar(const char cmd[], view_t *view,
 		MacroFlags flags);
 static int output_to_preview(view_t *view, const char cmd[], MacroFlags flags);
-static void run_in_split(const view_t *view, const char cmd[], int vert_split);
+static void run_in_split(const view_t *view, const char cmd[], int vert_split,
+		int pause);
 static void path_handler(const char line[], void *arg);
 static void line_handler(const char line[], void *arg);
 
@@ -183,19 +181,21 @@ handle_file(view_t *view, FileHandleExec exec, FileHandleLink follow)
 	int selected_entry = (curr->marked && (!user_selection || curr->selected));
 	if(!selected_entry && could_enter_entry)
 	{
-		int dir_like_entry = (is_dir(full_path) || is_unc_root(view->curr_dir));
+		int dir_like_entry = (fentry_is_dir(curr) || is_unc_root(view->curr_dir));
 		if(dir_like_entry)
 		{
-			enter_dir(view);
+			(void)rn_enter_dir(view);
 			return;
 		}
 	}
 
-	int runnable = is_runnable(view, full_path, curr->type,
-			follow != FHL_NO_FOLLOW);
-	int executable = is_executable(full_path, curr, exec == FHE_NO_RUN, runnable);
+	int multiselect = is_multiselect(view);
+	int runnable = is_runnable(full_path, curr->type, follow != FHL_NO_FOLLOW);
+	int executable = (exec != FHE_NO_RUN)
+	              && cfg.auto_execute
+	              && is_executable(full_path, curr, runnable);
 
-	if(stats_file_choose_action_set() && (executable || runnable))
+	if(stats_file_choose_action_set() && (multiselect || runnable || executable))
 	{
 		/* Reuse marking second time. */
 		view->pending_marking = 1;
@@ -207,7 +207,7 @@ handle_file(view_t *view, FileHandleExec exec, FileHandleLink follow)
 	{
 		execute_file(full_path, exec == FHE_ELEVATE_AND_RUN);
 	}
-	else if(runnable)
+	else if(multiselect || runnable)
 	{
 		run_selection(view, exec == FHE_NO_RUN);
 	}
@@ -217,25 +217,24 @@ handle_file(view_t *view, FileHandleExec exec, FileHandleLink follow)
 	}
 }
 
+/* Checks whether marking covers multiple files.  Returns non-zero if so. */
+static int
+is_multiselect(view_t *view)
+{
+	/* Checking for at least 2 files in the marking. */
+	dir_entry_t *entry = NULL;
+	return iter_marked_entries(view, &entry)
+	    && iter_marked_entries(view, &entry);
+}
+
 /* Returns non-zero if file can be executed or it's a link to a directory (it
  * can be entered), otherwise zero is returned. */
 static int
-is_runnable(view_t *view, const char full_path[], int type, int force_follow)
+is_runnable(const char full_path[], int type, int force_follow)
 {
-	int count = 0;
-	dir_entry_t *entry = NULL;
-	while(iter_marked_entries(view, &entry))
+	if(type == FT_LINK)
 	{
-		if(++count > 1)
-		{
-			return 1;
-		}
-	}
-
-	if(!force_follow && !cfg.follow_links && type == FT_LINK &&
-			get_symlink_type(full_path) != SLT_DIR)
-	{
-		return 1;
+		return (!force_follow && !cfg.follow_links);
 	}
 
 	if(type == FT_REG)
@@ -248,17 +247,19 @@ is_runnable(view_t *view, const char full_path[], int type, int force_follow)
 
 /* Returns non-zero if file can be executed, otherwise zero is returned. */
 static int
-is_executable(const char full_path[], const dir_entry_t *curr, int dont_execute,
-		int runnable)
+is_executable(const char full_path[], const dir_entry_t *curr, int runnable)
 {
-	int executable;
+	int executable = (curr->type == FT_EXEC);
 #ifndef _WIN32
-	executable = curr->type == FT_EXEC ||
-			(runnable && os_access(full_path, X_OK) == 0 && S_ISEXE(curr->mode));
-#else
-	executable = curr->type == FT_EXEC;
+	if(!executable)
+	{
+		/* XXX: why "runnable" is here?  Aren't the checks excessive? */
+		executable = runnable
+		          && (os_access(full_path, X_OK) == 0)
+		          && S_ISEXE(curr->mode);
+	}
 #endif
-	return executable && !dont_execute && cfg.auto_execute;
+	return executable;
 }
 
 #ifdef _WIN32
@@ -371,17 +372,18 @@ static void
 execute_file(const char full_path[], int elevate)
 {
 #ifndef _WIN32
-	char *const escaped = shell_like_escape(full_path, 0);
+	char *const escaped = shell_arg_escape(full_path, curr_stats.shell_type);
 	rn_shell(escaped, PAUSE_ALWAYS, 1, SHELL_BY_APP);
 	free(escaped);
 #else
-	char *const dquoted_full_path =
-		strdup(enclose_in_dquotes(full_path, curr_stats.shell_type));
+	char *copy = strdup(full_path);
+	internal_to_system_slashes(copy);
 
-	internal_to_system_slashes(dquoted_full_path);
-	run_win_executable(dquoted_full_path, elevate);
+	char *escaped = shell_arg_escape(copy, curr_stats.shell_type);
+	free(copy);
 
-	free(dquoted_full_path);
+	run_win_executable(escaped, elevate);
+	free(escaped);
 #endif
 }
 
@@ -401,7 +403,7 @@ run_selection(view_t *view, int dont_execute)
 	const char *common_prog_cmd = ft_get_program(typed_fname);
 	free(typed_fname);
 
-	int can_multi_run = (is_multi_run_compat(view, common_prog_cmd) != 0);
+	int can_multi_run = (is_multi_run_compat(common_prog_cmd) != 0);
 	int files_without_handler = (common_prog_cmd == NULL);
 	int identical_handlers = 1;
 	int nentries = 0;
@@ -428,7 +430,7 @@ run_selection(view_t *view, int dont_execute)
 			continue;
 		}
 
-		can_multi_run &= (is_multi_run_compat(view, entry_prog_cmd) != 0);
+		can_multi_run &= (is_multi_run_compat(entry_prog_cmd) != 0);
 		if(common_prog_cmd == NULL)
 		{
 			common_prog_cmd = entry_prog_cmd;
@@ -455,8 +457,7 @@ run_selection(view_t *view, int dont_execute)
 	}
 	else
 	{
-		show_error_msg("Run error", "Handlers of selected files are "
-		                            "incompatible.");
+		show_error_msg("Run error", "Handlers of selected files are incompatible.");
 	}
 }
 
@@ -467,7 +468,7 @@ run_with_defaults(view_t *view)
 {
 	if(get_current_entry(view)->type == FT_DIR)
 	{
-		enter_dir(view);
+		(void)rn_enter_dir(view);
 		return;
 	}
 
@@ -506,7 +507,7 @@ run_selection_separately(view_t *view, int dont_execute)
  * set of selected files.  Returns non-zero if so, otherwise zero is
  * returned. */
 static int
-is_multi_run_compat(view_t *view, const char prog_cmd[])
+is_multi_run_compat(const char prog_cmd[])
 {
 	size_t len;
 	if(prog_cmd == NULL)
@@ -516,11 +517,9 @@ is_multi_run_compat(view_t *view, const char prog_cmd[])
 	/* XXX: should the check be for " &" and not just "&"? */
 	if(prog_cmd[len - 1] != '&')
 		return 0;
-	if(strstr(prog_cmd, "%f") != NULL || strstr(prog_cmd, "%F") != NULL)
-		return 0;
-	if(strstr(prog_cmd, "%c") == NULL && strstr(prog_cmd, "%C") == NULL)
-		return 0;
-	return 1;
+
+	return ma_contains2(prog_cmd, MK_c, MK_C)
+	    && !ma_contains2(prog_cmd, MK_f, MK_F);
 }
 
 void
@@ -555,7 +554,7 @@ rn_open_with(view_t *view, const char prog_spec[], int dont_execute,
 	}
 	else if(strcmp(prog_spec, VIFM_PSEUDO_CMD) == 0)
 	{
-		enter_dir(view);
+		(void)rn_enter_dir(view);
 	}
 	else if(strchr(prog_spec, '%') != NULL)
 	{
@@ -581,7 +580,7 @@ run_explicit_prog(view_t *view, const char prog_spec[], int pause, int force_bg)
 	bg = !pause && (bg || force_bg);
 
 	int save_msg = 0;
-	if(rn_ext(view, cmd, prog_spec, flags, bg, &save_msg) != 0)
+	if(rn_ext(view, cmd, prog_spec, flags, pause, bg, &save_msg) != 0)
 	{
 		if(save_msg)
 		{
@@ -743,27 +742,29 @@ follow_link(view_t *view, int follow_dirs, int ultimate)
 	free(dir);
 }
 
-/* Handles opening of current entry of the view as a directory. */
-static void
-enter_dir(view_t *view)
+int
+rn_enter_dir(view_t *view)
 {
 	dir_entry_t *const curr = get_current_entry(view);
 
 	if(is_parent_dir(curr->name) && !curr->owns_origin)
 	{
 		rn_leave(view, 1);
-		return;
+		return 0;
 	}
 
 	char full_path[PATH_MAX + 1];
 	get_full_path_of(curr, sizeof(full_path), full_path);
-
-	if(cd_is_possible(full_path))
+	if(!cd_is_possible(full_path))
 	{
-		curr_stats.ch_pos = (cfg_ch_pos_on(CHPOS_ENTER) ? 1 : 0);
-		navigate_to(view, full_path);
-		curr_stats.ch_pos = 1;
+		return 1;
 	}
+
+	curr_stats.ch_pos = (cfg_ch_pos_on(CHPOS_ENTER) ? 1 : 0);
+	int result = navigate_to(view, full_path);
+	curr_stats.ch_pos = 1;
+
+	return result;
 }
 
 void
@@ -926,10 +927,7 @@ run_shell_finish(const char cmd[], const char final_cmd[], ShellPause pause,
 		stats_redraw_later();
 	}
 
-	if(curr_stats.load_stage > 0)
-	{
-		curs_set(0);
-	}
+	ui_set_cursor(/*visibility=*/0);
 
 	if(cmd == NULL)
 	{
@@ -970,7 +968,7 @@ setup_shellout_env(void)
 			return;
 	}
 
-	escaped_path = shell_like_escape(mount_file, 0);
+	escaped_path = shell_arg_escape(mount_file, curr_stats.shell_type);
 	cmd = format_str(term_multiplexer_fmt, FUSE_FILE_ENVVAR, escaped_path);
 	(void)vifm_system(cmd, SHELL_BY_APP);
 	free(cmd);
@@ -1042,7 +1040,7 @@ gen_term_multiplexer_cmd(const char cmd[], int pause, ShellRequester by)
 	title_arg = gen_term_multiplexer_title_arg(cmd);
 
 	raw_shell_cmd = format_str("%s%s", cmd, pause ? PAUSE_STR : "");
-	escaped_shell_cmd = shell_like_escape(raw_shell_cmd, 0);
+	escaped_shell_cmd = shell_arg_escape(raw_shell_cmd, curr_stats.shell_type);
 
 	const char *sh_flag = (by == SHELL_BY_USER ? cfg.shell_cmd_flag : "-c");
 
@@ -1050,7 +1048,7 @@ gen_term_multiplexer_cmd(const char cmd[], int pause, ShellRequester by)
 	{
 		char *const arg = format_str("%s %s %s", cfg.shell, sh_flag,
 				escaped_shell_cmd);
-		char *const escaped_arg = shell_like_escape(arg, 0);
+		char *const escaped_arg = shell_arg_escape(arg, curr_stats.shell_type);
 
 		shell_cmd = format_str("tmux new-window %s %s", title_arg, escaped_arg);
 
@@ -1115,7 +1113,7 @@ gen_term_multiplexer_title_arg(const char cmd[])
 	else
 	{
 		const char opt_c = (curr_stats.term_multiplexer == TM_SCREEN) ? 't' : 'n';
-		char *const escaped_title = shell_like_escape(title, 0);
+		char *const escaped_title = shell_arg_escape(title, curr_stats.shell_type);
 		title_arg = format_str("-%c %s", opt_c, escaped_title);
 		free(escaped_title);
 	}
@@ -1161,7 +1159,7 @@ gen_normal_cmd(const char cmd[], int pause)
 static void
 set_pwd_in_screen(const char path[])
 {
-	char *const escaped_dir = shell_like_escape(path, 0);
+	char *const escaped_dir = shell_arg_escape(path, curr_stats.shell_type);
 	char *const set_pwd = format_str("screen -X setenv PWD %s", escaped_dir);
 
 	(void)vifm_system(set_pwd, SHELL_BY_APP);
@@ -1218,7 +1216,7 @@ try_run_with_filetype(view_t *view, const assoc_records_t assocs,
 
 int
 rn_ext(view_t *view, const char cmd[], const char title[], MacroFlags flags,
-		int bg, int *save_msg)
+		int pause, int bg, int *save_msg)
 {
 	if(bg && (ma_flags_missing(flags, MF_NONE) &&
 	          ma_flags_missing(flags, MF_NO_TERM_MUX) &&
@@ -1286,14 +1284,14 @@ rn_ext(view_t *view, const char cmd[], const char title[], MacroFlags flags,
 	        curr_stats.term_multiplexer != TM_NONE)
 	{
 		const int vert_split = ma_flags_present(flags, MF_SPLIT_VERT);
-		run_in_split(view, cmd, vert_split);
+		run_in_split(view, cmd, vert_split, pause);
 	}
 	else if(ma_flags_present(flags, MF_CUSTOMVIEW_OUTPUT) ||
 	        ma_flags_present(flags, MF_VERYCUSTOMVIEW_OUTPUT) ||
 	        ma_flags_present(flags, MF_CUSTOMVIEW_IOUTPUT) ||
 	        ma_flags_present(flags, MF_VERYCUSTOMVIEW_IOUTPUT))
 	{
-		rn_for_flist(view, cmd, title, flags);
+		rn_for_flist(view, cmd, title, /*user_sh=*/1, flags);
 	}
 	else
 	{
@@ -1306,7 +1304,7 @@ rn_ext(view_t *view, const char cmd[], const char title[], MacroFlags flags,
 static void
 output_to_statusbar(const char cmd[], view_t *view, MacroFlags flags)
 {
-	FILE *file, *err;
+	FILE *file;
 	char buf[2048];
 	char *lines;
 	size_t len;
@@ -1314,7 +1312,8 @@ output_to_statusbar(const char cmd[], view_t *view, MacroFlags flags)
 
 	setup_shellout_env();
 	int error = 0;
-	if(bg_run_and_capture((char *)cmd, 1, input_tmp, &file, &err) == (pid_t)-1)
+	if(bg_run_and_capture((char *)cmd, 1, input_tmp, &file, /*err=*/NULL) ==
+			(pid_t)-1)
 	{
 		error = 1;
 	}
@@ -1333,6 +1332,9 @@ output_to_statusbar(const char cmd[], view_t *view, MacroFlags flags)
 
 	lines = NULL;
 	len = 0;
+	/* XXX: the loop can potentially never end if error pipe gets filled, might
+	 *      need to redirect stderr to /dev/null instead of opening and not using
+	 *      it. */
 	while(fgets(buf, sizeof(buf), file) == buf)
 	{
 		char *p;
@@ -1347,7 +1349,6 @@ output_to_statusbar(const char cmd[], view_t *view, MacroFlags flags)
 	}
 
 	fclose(file);
-	fclose(err);
 
 	ui_sb_msg((lines == NULL) ? "" : lines);
 	free(lines);
@@ -1370,11 +1371,19 @@ output_to_preview(view_t *view, const char cmd[], MacroFlags flags)
 /* Runs the cmd in a split window of terminal multiplexer.  Runs shell, if cmd
  * is NULL. */
 static void
-run_in_split(const view_t *view, const char cmd[], int vert_split)
+run_in_split(const view_t *view, const char cmd[], int vert_split, int pause)
 {
-	char *const escaped_cmd = (cmd == NULL)
-	                        ? strdup(cfg.shell)
-	                        : shell_like_escape(cmd, 0);
+	char *escaped_cmd;
+	if(cmd == NULL)
+	{
+		escaped_cmd = strdup(cfg.shell);
+	}
+	else
+	{
+		char *raw_shell_cmd = format_str("%s%s", cmd, pause ? PAUSE_STR : "");
+		escaped_cmd = shell_arg_escape(raw_shell_cmd, curr_stats.shell_type);
+		free(raw_shell_cmd);
+	}
 
 	setup_shellout_env();
 
@@ -1391,7 +1400,8 @@ run_in_split(const view_t *view, const char cmd[], int vert_split)
 
 		/* "eval" executes each argument as a separate argument, but escaping rules
 		 * are not exactly like in shell, so last command is run separately. */
-		char *const escaped_dir = shell_like_escape(flist_get_dir(view), 0);
+		char *const escaped_dir =
+			shell_arg_escape(flist_get_dir(view), curr_stats.shell_type);
 		if(vert_split)
 		{
 			snprintf(cmd, sizeof(cmd), "screen -X eval chdir\\ %s 'focus right' "
@@ -1440,7 +1450,7 @@ rn_start_bg_command(view_t *view, const char cmd[], MacroFlags flags)
 }
 
 int
-rn_for_flist(view_t *view, const char cmd[], const char title[],
+rn_for_flist(view_t *view, const char cmd[], const char title[], int user_sh,
 		MacroFlags flags)
 {
 	enum { MAX_TITLE_WIDTH = 80 };
@@ -1468,8 +1478,8 @@ rn_for_flist(view_t *view, const char cmd[], const char title[],
 	FILE *input_tmp = make_in_file(view, flags);
 
 	setup_shellout_env();
-	int error = (process_cmd_output("Loading custom view", cmd, input_tmp, 1,
-				interactive, &path_handler, view) != 0);
+	int error = (process_cmd_output("Loading custom view", cmd, input_tmp,
+				user_sh, interactive, &path_handler, view) != 0);
 	cleanup_shellout_env();
 
 	if(input_tmp != NULL)
@@ -1533,5 +1543,63 @@ line_handler(const char line[], void *arg)
 	list->nitems = add_to_string_array(&list->items, list->nitems, line);
 }
 
+int
+rn_cmd_exists(const char cmd[])
+{
+	if(vlua_handler_cmd(curr_stats.vlua, cmd))
+	{
+		return vlua_handler_present(curr_stats.vlua, cmd);
+	}
+
+	if(curr_stats.shell_type == ST_CMD || curr_stats.shell_type == ST_YORI)
+	{
+		/* This is a builtin command. */
+		if(stroscmp(cmd, "start") == 0)
+		{
+			return 1;
+		}
+	}
+
+	char path[PATH_MAX + 1];
+	if(rn_find_cmd(cmd, sizeof(path), path) == 0)
+	{
+		return executable_exists(path);
+	}
+
+	return 0;
+}
+
+int
+rn_find_cmd(const char cmd[], size_t path_len, char path[])
+{
+	if(starts_with(cmd, "!!"))
+	{
+		cmd += 2;
+	}
+
+	char *const expanded = replace_tilde(expand_envvars(cmd, EEF_NONE));
+	if(expanded == NULL)
+	{
+		return 1;
+	}
+
+	int failed;
+	if(contains_slash(cmd))
+	{
+		failed = 0;
+		if(copy_str(path, path_len, expanded) == path_len && path_len > 0 &&
+				path[path_len - 1] != '\0')
+		{
+			failed = 1;
+		}
+	}
+	else
+	{
+		failed = find_cmd_in_path(expanded, path_len, path);
+	}
+	free(expanded);
+	return failed;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
