@@ -863,7 +863,8 @@ qv_expand_viewer(view_t *view, const char viewer[], MacroFlags *flags)
 	char *result;
 	if(strchr(viewer, '%') == NULL)
 	{
-		char *escaped = shell_like_escape(get_current_file_name(view), 0);
+		char *escaped =
+			shell_arg_escape(get_current_file_name(view), curr_stats.shell_type);
 		result = format_str("%s %s", viewer, escaped);
 		free(escaped);
 	}
@@ -916,6 +917,8 @@ qv_cleanup_area(const preview_area_t *parea, const char cmd[])
 	}
 	else
 	{
+		/* XXX: this should probably get SHELL_BY_APP, otherwise preview command is
+		 *      called with SHELL_BY_APP, but clear command with SHELL_BY_USER. */
 		FILE *fp = read_cmd_output(expanded, /*preserve_stdin=*/0);
 		while(fgetc(fp) != EOF);
 		fclose(fp);
