@@ -260,9 +260,6 @@ clone_view(view_t *dst, view_t *side, const char path[], int clean)
 			path == NULL ? flist_get_dir(side) : path);
 
 	flist_init_view(dst);
-	/* This is for replace_dir_entries() below due to check in fentry_free(),
-	 * should adjust the check instead? */
-	dst->dir_entry[0].origin = side->curr_dir;
 
 	clone_local_options(side, dst, 1);
 	reset_local_options(dst);
@@ -405,8 +402,7 @@ tabs_goto_global(int idx)
 		return;
 	}
 
-	ui_qv_cleanup_if_needed();
-	modview_hide_graphics();
+	ui_hide_graphics();
 
 	global_tab_t *old_gtab = &gtabs[current_gtab];
 	global_tab_t *new_gtab = &gtabs[idx];
