@@ -18,7 +18,7 @@
 
 #include "vcache.h"
 
-#include <fcntl.h> /* F_GETFL O_NONBLOCK fcntl() */
+#include <fcntl.h> /* F_GETFL F_SETFL O_NONBLOCK fcntl() */
 
 #include <stdio.h> /* FILE */
 #include <stdlib.h> /* free() */
@@ -724,12 +724,13 @@ view_external(vcache_entry_t *centry, MacroFlags flags, const char **error)
 		bg_flags |= BJF_SUPPLY_INPUT;
 	}
 
-	if(ma_flags_present(flags, MF_KEEP_SESSION))
+	if(ma_flags_present(flags, MF_KEEP_IN_FG))
 	{
-		bg_flags |= BJF_KEEP_SESSION;
+		bg_flags |= BJF_KEEP_IN_FG;
 	}
 
-	centry->job = bg_run_external_job(centry->viewer, bg_flags);
+	centry->job =
+		bg_run_external_job(centry->viewer, bg_flags, /*descr=*/NULL, /*pwd=*/NULL);
 	if(centry->job == NULL)
 	{
 		*error = "Failed to start a viewer";
@@ -753,7 +754,7 @@ view_external(vcache_entry_t *centry, MacroFlags flags, const char **error)
 
 #ifndef _WIN32
 	/* Enable non-blocking read from output pipe.  On Windows we read the
-		* exact amount of data present in the stream. */
+	 * exact amount of data present in the stream. */
 	int fd = fileno(centry->job->output);
 	int file_flags = fcntl(fd, F_GETFL, 0);
 	fcntl(fd, F_SETFL, file_flags | O_NONBLOCK);
